@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 1999-2021. All Rights Reserved.
+ * Copyright Ericsson AB 1999-2023. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -302,15 +302,35 @@ struct m {
     Eterm exception;
 };
 
-static Eterm staging_epilogue(Process* c_p, int, Eterm res, int, struct m*, int, int);
-static void smp_code_ix_commiter(void*);
-
 static struct /* Protected by code_write_permission */
 {
     Process* stager;
-    ErtsThrPrgrLaterOp lop;
+    ErtsCodeBarrier barrier;
 } committer_state;
 
+static Eterm staging_epilogue(Process* c_p, int, Eterm res, int, struct m*, int, int);
+
+static void commit_code_ix(void *null)
+{
+    Process* p = committer_state.stager;
+
+    erts_commit_staging_code_ix();
+
+#ifdef DEBUG
+    committer_state.stager = NULL;
+#endif
+
+    erts_release_code_write_permission();
+
+    erts_proc_lock(p, ERTS_PROC_LOCK_STATUS);
+    if (!ERTS_PROC_IS_EXITING(p)) {
+        erts_resume(p, ERTS_PROC_LOCK_STATUS);
+    }
+    erts_proc_unlock(p, ERTS_PROC_LOCK_STATUS);
+
+    erts_proc_dec_refc(p);
+}
+
 static Eterm
 exception_list(Process* p, Eterm tag, struct m* mp, Sint exceptions)
 {
@@ -503,17 +523,19 @@ staging_epilogue(Process* c_p, int commit, Eterm res, int is_blocking,
 {    
     if (is_blocking || !commit)
     {
-	if (commit) {
-	    int i;
-	    erts_end_staging_code_ix();
-	    erts_commit_staging_code_ix();
+        if (commit) {
+            int i;
+            erts_end_staging_code_ix();
+            erts_commit_staging_code_ix();
 
-	    for (i=0; i < nmods; i++) {
-		if (mods[i].modp->curr.code_hdr
-                    && mods[i].exception != am_on_load) {
-		    set_default_trace_pattern(mods[i].module);
-		}
-	    }
+            for (i=0; i < nmods; i++) {
+                if (mods[i].modp->curr.code_hdr
+                     && mods[i].exception != am_on_load) {
+                    set_default_trace_pattern(mods[i].module);
+                }
+            }
+
+            erts_blocking_code_barrier();
 	}
 	else {
 	    erts_abort_staging_code_ix();
@@ -542,37 +564,19 @@ staging_epilogue(Process* c_p, int commit, Eterm res, int is_blocking,
 	 */
 	ASSERT(committer_state.stager == NULL);
 	committer_state.stager = c_p;
-	erts_schedule_thr_prgr_later_op(smp_code_ix_commiter, NULL, &committer_state.lop);
-	erts_proc_inc_refc(c_p);
-	erts_suspend(c_p, ERTS_PROC_LOCK_MAIN, NULL);
-	/*
-	 * smp_code_ix_commiter() will do the rest "later"
-	 * and resume this process to return 'res'.  
-	 */
-	ERTS_BIF_YIELD_RETURN(c_p, res);
+
+        erts_schedule_code_barrier(&committer_state.barrier,
+                                   commit_code_ix, NULL);
+
+        erts_proc_inc_refc(c_p);
+        erts_suspend(c_p, ERTS_PROC_LOCK_MAIN, NULL);
+
+        /* commit_code_ix(NULL) will resume us when everything is finished, at
+         * which point we'll return `res`. */
+        ERTS_BIF_YIELD_RETURN(c_p, res);
     }
 }
 
-
-static void smp_code_ix_commiter(void* null)
-{
-    Process* p = committer_state.stager;
-
-    erts_commit_staging_code_ix();
-#ifdef DEBUG
-    committer_state.stager = NULL;
-#endif
-    erts_release_code_write_permission();
-    erts_proc_lock(p, ERTS_PROC_LOCK_STATUS);
-    if (!ERTS_PROC_IS_EXITING(p)) {
-	erts_resume(p, ERTS_PROC_LOCK_STATUS);
-    }
-    erts_proc_unlock(p, ERTS_PROC_LOCK_STATUS);
-    erts_proc_dec_refc(p);
-}
-
-
-
 BIF_RETTYPE
 check_old_code_1(BIF_ALIST_1)
 {
@@ -974,13 +978,40 @@ set_default_trace_pattern(Eterm module)
     }
 }
 
+int
+erts_check_copy_literals_gc_need_max_reds(Process *c_p)
+{
+    Uint64 words, reds;
+
+    /*
+     * Calculate maximum amount of words that needs
+     * to be scanned...
+     */
+    words = 1; /* fvalue */
+    words += c_p->hend - c_p->stop; /* stack */
+    words += c_p->htop - c_p->heap; /* new heap */
+    if (c_p->abandoned_heap)
+        words += c_p->heap_sz; /* abandoned heap */
+    words += c_p->old_htop - c_p->old_heap; /* old heap */
+    if (c_p->dictionary) {
+	Eterm* start = ERTS_PD_START(c_p->dictionary);
+	Eterm* end = start + ERTS_PD_SIZE(c_p->dictionary);
+
+        words += end - start; /* dictionary */
+    }
+    words += c_p->mbuf_sz; /* heap and message fragments */
+
+    /* Convert to reductions... */
+    reds = ((words - 1)/ERTS_CLA_SCAN_WORDS_PER_RED) + 1;
+    if (reds > CONTEXT_REDS)
+        return CONTEXT_REDS+1;
+    return (int) reds;
+}
+
 int
 erts_check_copy_literals_gc_need(Process *c_p, int *redsp,
                                  char *literals, Uint lit_bsize)
 {
-    /*
-     * TODO: Implement yielding support!
-     */
     ErlHeapFragment *hfrag;
     ErtsMessage *mfp;
     Uint64 scanned = 0;
@@ -1529,22 +1560,30 @@ erts_literal_area_collector_send_copy_request_3(BIF_ALIST_3)
 
     req_id = TUPLE3(&tmp_heap[0], BIF_ARG_2, BIF_ARG_3, BIF_ARG_1);
 
-    if (BIF_ARG_3 == am_false) {
+    switch (BIF_ARG_3) {
+
+    case am_init:
         /*
-         * Will handle signal queue and check if GC if needed. If
-         * GC is needed operation will be continued by a GC (below).
+         * Will handle signal queue and if possible check if GC if needed.
+         * If GC is needed or needs to be checked the operation will be
+         * restarted later in the 'check_gc' or 'need_gc' case below...
          */
         erts_proc_sig_send_cla_request(BIF_P, BIF_ARG_1, req_id);
-    }
-    else if (BIF_ARG_3 == am_true) {
+        break;
+
+    case am_check_gc:
+    case am_need_gc:
         /*
-         * Will perform a literal GC. Note that this assumes that
-         * signal queue already has been handled...
+         * Will check and/or perform a literal GC. Note that this assumes that
+         * signal queue already has been handled by 'init' case above...
          */
-        erts_schedule_cla_gc(BIF_P, BIF_ARG_1, req_id);
-    }
-    else
+        erts_schedule_cla_gc(BIF_P, BIF_ARG_1, req_id,
+                             BIF_ARG_3 == am_check_gc);
+        break;
+
+    default:
         BIF_ERROR(BIF_P, BADARG);
+    }
 
     BIF_RET(am_ok);
 }
