@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2020-2022. All Rights Reserved.
+ * Copyright Ericsson AB 2020-2023. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -698,8 +698,9 @@ void BeamModuleAssembler::emit_i_rem_div(const ArgLabel &Fail,
                                          const ArgRegister &Remainder,
                                          const ArgRegister &Quotient) {
     static const ErtsCodeMFA bif_mfa = {am_erlang, am_rem, 2};
+    bool need_rem = Quotient != Remainder;
 
-    emit_div_rem(Fail, LHS, RHS, &bif_mfa, Quotient, Remainder, true, true);
+    emit_div_rem(Fail, LHS, RHS, &bif_mfa, Quotient, Remainder, true, need_rem);
 }
 
 void BeamModuleAssembler::emit_i_div_rem(const ArgLabel &Fail,
@@ -709,8 +710,9 @@ void BeamModuleAssembler::emit_i_div_rem(const ArgLabel &Fail,
                                          const ArgRegister &Quotient,
                                          const ArgRegister &Remainder) {
     static const ErtsCodeMFA bif_mfa = {am_erlang, am_div, 2};
+    bool need_div = Quotient != Remainder;
 
-    emit_div_rem(Fail, LHS, RHS, &bif_mfa, Quotient, Remainder, true, true);
+    emit_div_rem(Fail, LHS, RHS, &bif_mfa, Quotient, Remainder, need_div, true);
 }
 
 void BeamModuleAssembler::emit_i_int_div(const ArgLabel &Fail,
@@ -991,6 +993,7 @@ void BeamModuleAssembler::emit_i_bxor(const ArgLabel &Fail,
 
     if (Fail.get() != 0) {
         emit_enter_runtime(Live.get());
+        a.mov(ARG1, c_p);
         runtime_call<3>(erts_bxor);
         emit_leave_runtime(Live.get());
         emit_branch_if_not_value(ARG1, resolve_beam_label(Fail, dispUnknown));
