@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %% 
-%% Copyright Ericsson AB 2002-2021. All Rights Reserved.
+%% Copyright Ericsson AB 2002-2023. All Rights Reserved.
 %% 
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -44,18 +44,18 @@
 
 -record(timers, 
         {
-          request_timers = [] :: [reference()],
-          queue_timer         :: reference() | 'undefined'
+          request_timers = [] :: [reference() | {reference(), term()}],
+          queue_timer         :: reference() | undefined
          }).
 
 -record(state, 
         {
-          request                   :: request() | 'undefined',
-          session                   :: session() | 'undefined',
-          status_line,               % {Version, StatusCode, ReasonPharse}
-          headers                   :: http_response_h() | 'undefined',
-          body                      :: binary() | 'undefined',
-          mfa,                       % {Module, Function, Args}
+          request                   :: request() | undefined,
+          session                   :: session() | undefined,
+          status_line               :: tuple()   | undefined,     % {Version, StatusCode, ReasonPharse}
+          headers                   :: http_response_h() | undefined,
+          body                      :: binary() | undefined,
+          mfa                       :: {atom(), atom(), term()} | undefined, % {Module, Function, Args}
           pipeline = queue:new()    :: queue:queue(),
           keep_alive = queue:new()  :: queue:queue(),
           status                    :: undefined | new | pipeline | keep_alive | close | {ssl_tunnel, request()},
@@ -65,7 +65,7 @@
           options                   :: options(),
           timers = #timers{}        :: #timers{},
           profile_name              :: atom(), % id of httpc_manager process.
-          once = inactive           :: 'inactive' | 'once'
+          once = inactive           :: inactive | once
          }).
 
 
@@ -515,7 +515,7 @@ do_handle_info({Proto, _Socket, Data},
 	    handle_http_msg(Result, State); 
 	{_, whole_body, _} when Method =:= head ->
 	    handle_response(State#state{body = <<>>}); 
-	{Module, whole_body, [Body, Length]} ->
+	{Module, whole_body, [Body, Length]} when is_binary(Body)->
 	    {_, Code, _} = StatusLine,
 	    {Streamed, NewBody, NewRequest} = stream(Body, Request, Code),
 	    %% When we stream we will not keep the already
@@ -525,7 +525,7 @@ do_handle_info({Proto, _Socket, Data},
 		    false ->
 			Length;
 		    true ->
-			Length - size(Body)                     
+			Length - byte_size(Body)
 		end,
 	    
 	    NewState = next_body_chunk(State, Code),
@@ -605,7 +605,7 @@ do_handle_info({Proto, Socket, Data},
 %% whole body is now sent instead of sending a length
 %% indicator. In this case the length indicator will be
 %% -1.
-do_handle_info({Info, _}, State = #state{mfa = {_, whole_body, Args}})
+do_handle_info({Info, _}, #state{mfa = {_, whole_body, Args}} = State)
   when Info =:= tcp_closed orelse
        Info =:= ssl_closed ->
     case lists:last(Args) of
@@ -616,25 +616,18 @@ do_handle_info({Info, _}, State = #state{mfa = {_, whole_body, Args}})
     end;
 
 %%% Server closes idle pipeline
-do_handle_info({tcp_closed, _}, State = #state{request = undefined}) ->
+do_handle_info({tcp_closed, _}, #state{request = undefined} = State) ->
     {stop, normal, State};
-do_handle_info({ssl_closed, _}, State = #state{request = undefined}) ->
+do_handle_info({ssl_closed, _}, #state{request = undefined} = State) ->
     {stop, normal, State};
 
 %%% Error cases
-do_handle_info({tcp_closed, _}, #state{session = Session0} = State) ->
-    Socket  = Session0#session.socket,
+do_handle_info({Closed, _}, #state{session = #session{socket = Socket}=Session0} = State)
+  when Closed == tcp_closed; Closed == ssl_closed ->
     Session = Session0#session{socket = {remote_close, Socket}},
     %% {stop, session_remotly_closed, State};
     {stop, normal, State#state{session = Session}};
-do_handle_info({ssl_closed, _}, #state{session = Session0} = State) ->
-    Socket  = Session0#session.socket,
-    Session = Session0#session{socket = {remote_close, Socket}},
-    %% {stop, session_remotly_closed, State};
-    {stop, normal, State#state{session = Session}};
-do_handle_info({tcp_error, _, _} = Reason, State) ->
-    {stop, Reason, State};
-do_handle_info({ssl_error, _, _} = Reason, State) ->
+do_handle_info({Error, _, _} = Reason, State) when Error == tcp_error; Error == ssl_error ->
     {stop, Reason, State};
 
 %% Timeouts
@@ -675,7 +668,7 @@ do_handle_info({timeout, RequestId},
                                   keep_alive = KeepAlive}}
     end;
 
-do_handle_info(timeout_queue, State = #state{request = undefined}) ->
+do_handle_info(timeout_queue, #state{request = undefined} = State) ->
     {stop, normal, State};
 
 %% Timing was such as the queue_timeout was not canceled!
@@ -685,12 +678,12 @@ do_handle_info(timeout_queue, #state{timers = Timers} = State) ->
 
 %% Setting up the connection to the server somehow failed. 
 do_handle_info({init_error, Reason, ClientErrMsg},
-            State = #state{request = Request}) ->
+               #state{request = Request} = State) ->
     NewState = answer_request(Request, ClientErrMsg, State),
     {stop, {shutdown, Reason}, NewState};
 
 %%% httpc_manager process dies. 
-do_handle_info({'EXIT', _, _}, State = #state{request = undefined}) ->
+do_handle_info({'EXIT', _, _}, #state{request = undefined} = State) ->
     {stop, normal, State};
 %%Try to finish the current request anyway,
 %% there is a fairly high probability that it can be done successfully.
@@ -878,7 +871,7 @@ connect_and_send_upgrade_request(Address, Request, #state{options = Options0} =
     end.
 
 handler_info(#state{request     = Request, 
-		    session     = Session, 
+		    session     = #session{socket = Socket}=Session,
 		    status_line = _StatusLine, 
 		    pipeline    = Pipeline, 
 		    keep_alive  = KeepAlive, 
@@ -906,8 +899,7 @@ handler_info(#state{request     = Request,
 			  queue:len(KeepAlive)
 		  end,
     Scheme     = Session#session.scheme, 
-    Socket     = Session#session.socket, 
-    SocketType = Session#session.socket_type, 
+    SocketType = Session#session.socket_type,
 
     SocketOpts  = http_transport:getopts(SocketType, Socket), 
     SocketStats = http_transport:getstat(SocketType, Socket), 
@@ -962,7 +954,7 @@ handle_http_body(_, #state{status = {ssl_tunnel, _},
 
 handle_http_body(_, #state{status = {ssl_tunnel, Request},
 			   status_line = StatusLine} = State) ->
-    ClientErrMsg = httpc_response:error(Request,{could_no_establish_ssh_tunnel, StatusLine}),
+    ClientErrMsg = httpc_response:error(Request,{could_not_establish_ssl_tunnel, StatusLine}),
     NewState     = answer_request(Request, ClientErrMsg, State),
     {stop, normal, NewState};
 
@@ -983,11 +975,10 @@ handle_http_body(Body, #state{headers = Headers,
 handle_http_body(_Body, #state{request = #request{method = head}} = State) ->
     handle_response(State#state{body = <<>>});
 
-handle_http_body(Body, #state{headers       = Headers, 
+handle_http_body(Body, #state{headers       = #http_response_h{'transfer-encoding' = TransferEnc}=Headers,
 			      max_body_size = MaxBodySize,
 			      status_line   = {_,Code, _},
 			      request       = Request} = State) ->
-    TransferEnc = Headers#http_response_h.'transfer-encoding',
     case case_insensitive_header(TransferEnc) of
         "chunked" ->
 	    try http_chunk:decode(Body, State#state.max_body_size, 
@@ -1064,7 +1055,7 @@ handle_response(#state{status = Status0} = State0) when Status0 =/= new ->
     handle_cookies(Headers, Request, Options, ProfileName), 
     case httpc_response:result({StatusLine, Headers, Body}, Request) of
 	%% 100-continue
-	continue -> 
+	continue ->
 	    %% Send request body
 	    {_, RequestBody} = Request#request.content,
 	    send_raw(Session, RequestBody),
@@ -1080,7 +1071,7 @@ handle_response(#state{status = Status0} = State0) when Status0 =/= new ->
 
 	%% Ignore unexpected 100-continue response and receive the
 	%% actual response that the server will send right away. 
-	{ignore, Data} -> 
+	{ignore, Data} ->
 	    Relaxed = (Request#request.settings)#http_options.relaxed,
 	    MFA     = {httpc_response, parse,
 		       [State#state.max_header_size, Relaxed]}, 
@@ -1116,8 +1107,7 @@ handle_cookies(_,_, #options{cookies = disabled}, _) ->
 %% so the user will have to call a store command.
 handle_cookies(_,_, #options{cookies = verify}, _) ->
     ok;
-handle_cookies(Headers, Request, #options{cookies = enabled}, ProfileName) ->
-    {Host, _ } = Request#request.address,
+handle_cookies(Headers, #request{address = {Host, _}}=Request, #options{cookies = enabled}, ProfileName) ->
     Cookies = httpc_cookie:cookies(Headers#http_response_h.other, 
 				  Request#request.path, Host),
     httpc_manager:store_cookies(Cookies, Request#request.address,
@@ -1135,7 +1125,7 @@ handle_queue(#state{status = pipeline} = State, Data) ->
     handle_pipeline(State, Data).
 
 handle_pipeline(#state{status       = pipeline, 
-		       session      = Session,
+		       session      = #session{}=Session,
 		       profile_name = ProfileName,
 		       options      = #options{pipeline_timeout = TimeOut}} = State,
 		Data) ->
@@ -1637,14 +1627,14 @@ host_header(_, URI) ->
 tls_upgrade(#state{status = 
 		       {ssl_tunnel, 
 			#request{settings = 
-				     #http_options{ssl = {_, TLSOptions0} = SocketType},
+				     #http_options{ssl = {_, TLSOptions0} = SocketType,
+						   connect_timeout = ConnectTimeout},
 				     address = {Host, _} = Address} = Request},
 		   session = #session{socket = TCPSocket} = Session0,
 		   options = Options} = State) ->
 
     TLSOptions = maybe_add_sni(Host, TLSOptions0),
-
-    case ssl:connect(TCPSocket, TLSOptions) of
+    case ssl:connect(TCPSocket, TLSOptions, ConnectTimeout) of
 	{ok, TLSSocket} ->
 	    ClientClose = httpc_request:is_client_closing(Request#request.headers),
 	    SessionType = httpc_manager:session_type(Options),
