@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %% 
-%% Copyright Ericsson AB 1996-2018. All Rights Reserved.
+%% Copyright Ericsson AB 1996-2023. All Rights Reserved.
 %% 
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -59,6 +59,8 @@
 	 ext_pending = []     % [{ext,From}]
 	}).
 
+-define(EXT_MEM_MAP, memsup_ext_memory_type_map__).
+
 %%----------------------------------------------------------------------
 %% API
 %%----------------------------------------------------------------------
@@ -189,6 +191,7 @@ init([]) ->
 		   _ ->
 		       exit({unsupported_os, OS})
 	       end,
+
     Pid = if
 	      PortMode ->
 		  spawn_link(fun() -> port_init() end);
@@ -203,6 +206,21 @@ init([]) ->
     SysMem = os_mon:get_env(memsup, system_memory_high_watermark),
     ProcMem = os_mon:get_env(memsup, process_memory_high_watermark),
 
+    persistent_term:put(?EXT_MEM_MAP,
+                        #{?MEM_SYSTEM_TOTAL => system_total_memory,
+                          ?MEM_TOTAL => total_memory,
+                          ?MEM_AVAIL => available_memory,
+                          ?MEM_FREE => free_memory,
+                          ?MEM_BUFFERS => buffered_memory,
+                          ?MEM_CACHED => cached_memory,
+                          %% When included cached_x is same as cached...
+                          ?MEM_CACHED_X => cached_memory, 
+                          ?MEM_SHARED => shared_memory,
+                          ?MEM_LARGEST_FREE => largest_free,
+                          ?MEM_NUMBER_OF_FREE => number_of_free,
+                          ?SWAP_TOTAL => total_swap,
+                          ?SWAP_FREE => free_swap}),
+
     %% Initiate first data collection
     self() ! time_to_collect,
 
@@ -557,7 +575,8 @@ get_os_wordsize_with_uname() ->
 	"ppc64"   -> 64;
 	"ppc64le" -> 64;
 	"s390x"   -> 64;
-        "aarch64" -> 64;
+        "aarch64" -> 64;                        %Linux
+        "arm64"   -> 64;                        %macOS
 	_         -> 32
     end.
 
@@ -632,6 +651,10 @@ port_init() ->
 start_portprogram() ->
     os_mon:open_port("memsup",[{packet,1}]).
 
+port_shutdown(Port) ->
+    Port ! {self(), {command, [?EXIT]}},
+    port_close(Port).
+
 %% The connected process loops are a bit awkward (several different
 %% functions doing almost the same thing) as
 %%   a) strategies for receiving regular memory data and extensive
@@ -647,7 +670,7 @@ port_idle(Port) ->
 	    get_memory_usage(Port, undefined, Memsup);
 	{Memsup, collect_ext_sys} ->
 	    Port ! {self(), {command, [?SHOW_SYSTEM_MEM]}},
-	    get_ext_memory_usage(Port, [], Memsup);
+	    get_ext_memory_usage(Port, #{}, Memsup);
 	cancel ->
 	    %% Received after reply already has been delivered...
 	    port_idle(Port);
@@ -655,13 +678,13 @@ port_idle(Port) ->
 	    %% Received after reply already has been delivered...
 	    port_idle(Port);
 	close ->
-	    port_close(Port);
+	    port_shutdown(Port);
 	{Port, {data, Data}} ->
 	    exit({port_error, Data});
 	{'EXIT', Port, Reason} ->
 	    exit({port_died, Reason});
 	{'EXIT', _Memsup, _Reason} ->
-	    port_close(Port)
+	    port_shutdown(Port)
     end.
 
 get_memory_usage(Port, Alloc, Memsup) ->
@@ -675,11 +698,11 @@ get_memory_usage(Port, Alloc, Memsup) ->
 	cancel ->
 	    get_memory_usage_cancelled(Port, Alloc);
 	close ->
-	    port_close(Port);
+	    port_shutdown(Port);
 	{'EXIT', Port, Reason} ->
 	    exit({port_died, Reason});
 	{'EXIT', _Memsup, _Reason} ->
-	    port_close(Port)
+	    port_shutdown(Port)
     end.
 get_memory_usage_cancelled(Port, Alloc) ->
     receive
@@ -688,101 +711,84 @@ get_memory_usage_cancelled(Port, Alloc) ->
 	{Port, {data, _Data}} ->
 	    port_idle(Port);
 	close ->
-	    port_close(Port);
+	    port_shutdown(Port);
 	{'EXIT', Port, Reason} ->
 	    exit({port_died, Reason});
 	{'EXIT', _Memsup, _Reason} ->
-	    port_close(Port)
+	    port_shutdown(Port)
+    end.
+
+tag2atag(Port, Tag) ->
+    Tab = persistent_term:get(?EXT_MEM_MAP),
+    try
+        maps:get(Tag, Tab)
+    catch
+        error:_ ->
+            exit({memsup_port_error, {Port,[Tag]}})
     end.
 
 get_ext_memory_usage(Port, Accum, Memsup) ->
-    Tab = [
-	    {?MEM_SYSTEM_TOTAL,   system_total_memory},
-	    {?MEM_TOTAL,          total_memory},
-	    {?MEM_FREE,           free_memory},
-	    {?MEM_BUFFERS,        buffered_memory},
-	    {?MEM_CACHED,         cached_memory},
-	    {?MEM_SHARED,         shared_memory},
-	    {?MEM_LARGEST_FREE,   largest_free},
-	    {?MEM_NUMBER_OF_FREE, number_of_free},
-	    {?SWAP_TOTAL,	  total_swap},
-	    {?SWAP_FREE,	  free_swap}
-	],
     receive
 	{Port, {data, [?SHOW_SYSTEM_MEM_END]}} ->
-	    Memsup ! {collected_ext_sys, Accum},
+	    Memsup ! {collected_ext_sys, maps:to_list(Accum)},
 	    port_idle(Port);
 	{Port, {data, [Tag]}} ->
-	    case lists:keysearch(Tag, 1, Tab) of
-		{value, {Tag, ATag}} ->
-		    get_ext_memory_usage(ATag, Port, Accum, Memsup);
-		_ ->
-		    exit({memsup_port_error, {Port,[Tag]}})
-	    end;
+            get_ext_memory_usage(tag2atag(Port, Tag), Port, Accum,
+                                 Memsup);
 	ext_cancel ->
 	    get_ext_memory_usage_cancelled(Port);
 	close ->
-	    port_close(Port);
+	    port_shutdown(Port);
 	{'EXIT', Port, Reason} ->
 	    exit({port_died, Reason});
 	{'EXIT', _Memsup, _Reason} ->
-	    port_close(Port)
+	    port_shutdown(Port)
     end.
 get_ext_memory_usage_cancelled(Port) ->
-    Tab = [
-	    {?MEM_SYSTEM_TOTAL,   system_total_memory},
-	    {?MEM_TOTAL,          total_memory},
-	    {?MEM_FREE,           free_memory},
-	    {?MEM_BUFFERS,        buffered_memory},
-	    {?MEM_CACHED,         cached_memory},
-	    {?MEM_SHARED,         shared_memory},
-	    {?MEM_LARGEST_FREE,   largest_free},
-	    {?MEM_NUMBER_OF_FREE, number_of_free},
-	    {?SWAP_TOTAL,	  total_swap},
-	    {?SWAP_FREE,	  free_swap}
-	],
     receive
 	{Port, {data, [?SHOW_SYSTEM_MEM_END]}} ->
 	    port_idle(Port);
 	{Port, {data, [Tag]}} ->
-	    case lists:keysearch(Tag, 1, Tab) of
-		{value, {Tag, ATag}} ->
-		    get_ext_memory_usage_cancelled(ATag, Port);
-		_ ->
-		    exit({memsup_port_error, {Port,[Tag]}})
-	    end;
+            get_ext_memory_usage_cancelled(tag2atag(Port, Tag),
+                                           Port);
 	close ->
-	    port_close(Port);
+	    port_shutdown(Port);
 	{'EXIT', Port, Reason} ->
 	    exit({port_died, Reason});
 	{'EXIT', _Memsup, _Reason} ->
-	    port_close(Port)
+	    port_shutdown(Port)
     end.
 
 get_ext_memory_usage(ATag, Port, Accum0, Memsup) ->
     receive
 	{Port, {data, Data}} ->
-	    Accum = [{ATag,erlang:list_to_integer(Data, 16)}|Accum0],
+            Value = erlang:list_to_integer(Data, 16),
+            Accum = case maps:get(ATag, Accum0, undefined) of
+                        undefined ->
+                            maps:put(ATag, Value, Accum0);
+                        PrevValue ->
+                            maps:put(ATag, Value + PrevValue, Accum0)
+                    end,
 	    get_ext_memory_usage(Port, Accum, Memsup);
 	cancel ->
 	    get_ext_memory_usage_cancelled(ATag, Port);
 	close ->
-	    port_close(Port);
+	    port_shutdown(Port);
 	{'EXIT', Port, Reason} ->
 	    exit({port_died, Reason});
 	{'EXIT', _Memsup, _Reason} ->
-	    port_close(Port)
+	    port_shutdown(Port)
     end.
 get_ext_memory_usage_cancelled(_ATag, Port) ->
     receive
 	{Port, {data, _Data}} ->
 	    get_ext_memory_usage_cancelled(Port);
 	close ->
-	    port_close(Port);
+	    port_shutdown(Port);
 	{'EXIT', Port, Reason} ->
 	    exit({port_died, Reason});
 	{'EXIT', _Memsup, _Reason} ->
-	    port_close(Port)
+	    port_shutdown(Port)
     end.
 
 %%--Collect process data------------------------------------------------
